/* AOLbuffer source
 *
 * Copyright (C) 1999, 2000 Brent A. Fulgham <bfulgham@debian.org>
 * http://www.debian.org
 *
 * See also http://www.aolserver.com for details on the AOLserver
 * web server software
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 * 
 */

/*
 * The contents of this source file are based heavily on the following
 * sources:
 *
 * (*) Rob Mayoff's "tclpages" source code
 * (*) Lele Gaifax's "PyApache" source code
 * (*) Examples from the Python source distribution (especially the
 *     chapters on extended and embedding python)
 *
 * This source file is just the first step in Python world domination.
 * If this unit is successful, it will mean the begining of my secret
 * plot to replace the AOLserver Tcl base with Python.  Mwahahahaha.
 *
 * By the way, I tend to comment verbosely, so sorry if you are printing
 * this thing out.  But if I don't make notations here, I forget why I
 * did things and that just makes life harder in general.  Besides,
 * you're not paying for this so just enjoy. :-)
 * 
 */

#include "AOLbuffer.hh"
#include <algo.h>
#define CONN_CLOSED(conn) ((conn)->flags & NS_CONN_CLOSED)

//----------------------------------------------------------------------------
//
//  ParseHeaders(const char* str, int size)
//
//  This function gets called from the AOLbuffer::write() method until the
//  end-of-header condition (two consecutive newline characters) is met.
//
//  Splitting the input STR into '\n' separated lines, recognizes the
//  header and updates various tables.
//
//  Function returns the number of characters consumed, -1 on errors, and
//  -2 when there is a valid "Location" header.
//
//----------------------------------------------------------------------------
int
AOLbuffer::ParseHeaders(const char *str, int size)
{
    if (headersSent) {
        return EOF;
    }

    // Shortcut for common Python usage: when it executes "print 'blah
    // blah' ", it actually does write the string 'blah blah' to the
    // output stream, then it writes a single newline character.  If
    // so, maybe we can save a realloc and an empty scan...
    int consumed = 0;
    if (*str == '\n'
        || ((size > 1) && (*str == '\r') && (*(str + 1) == '\n'))) {
        if (!httpHeaders->empty()) {
            // If there is an incomplete header, consider it done.
            if (httpHeaders->size() > 2) {
                // Avoid "\n\0" case
                if (HandleHeader()) {
                    return NS_ERROR;
                }
            }

            // Headers parsed -- clear them out.
            httpHeaders->clear();

            // If there is just the newline...
            if ((size == 1) || ((size == 2) && (*str == '\r'))) {
                Ns_Log(Debug, "PyWX: Headers Sent!");
                Ns_ConnFlushHeaders(connection, HTTP_OK);
                headersSent = true;
                return size;
            }

            // ...nope.  Something more...
            consumed = ((*str == '\n') ? 1 : 2);
            str += consumed;
            size -= consumed;
        }
        else {
            // Otherwise, we're done.
            if (redirected) {
                if (redirected == REDIRECT_INTERNAL) {
                    Ns_ConnRedirect(connection, connection->request->url);
                }

                return size;
            }

            // Send Header Data
            Ns_ConnFlushHeaders(connection, HTTP_OK);
            headersSent = true;

            // Our caller will eventually write the rest of the string
            return ((*str == '\n') ? 1 : 2);
        }
    }

    //-----------------------------------------------------------------------
    //
    //  Begin *real* Header processing
    //
    //-----------------------------------------------------------------------

    int endOfHeader = 0;

    char *lbeg = (char *) &str[0];
    char *lend = strchr(lbeg, '\n');

    if (lend == NULL) {
        lend = (char *) &str[size];
    }

    int llen = lend - lbeg;

    while (!endOfHeader && lend != NULL) {
        // First, append our new stuff to our buffer
        httpHeaders->clear();
        copy(lbeg, lend, back_inserter(*httpHeaders));
        httpHeaders->push_back('\0');

        // Ok, we found EOL, but is it a Messy DOS-ish CR/NL pair instead?
        if (llen != 0) {
            if ((llen > 1) && (*(lend - 1) == '\r')) {
                llen -= 2;
                *(lend - 1) = '\0';	// We now have "\0\n"
            }
            else {
                llen--;
                *lend = '\0';
            }
        }

        // Is anything left after the above manipulations?
        if (llen == 0) {
            endOfHeader = 1;

            // Done!  No more headers.  Now, send them back, then
            // finish writing the string to the client.
            int status = Ns_ConnResponseStatus(connection);

            if ((status == HTTP_OK) && (redirected)) {
                if (redirected == REDIRECT_INTERNAL) {
                    Ns_ConnRedirect(connection, connection->request->url);
                }
                return size;
            }

            // Send header data
            Ns_ConnFlushHeaders(connection, HTTP_OK);
            headersSent = true;
            lbeg = lend + 1;
            break;
        }
        else {
            if (HandleHeader()) {
                return NS_ERROR;
            }
        }
        lbeg = lend + 1;		// Puts us on the '\n' of an '\0\n'
        lend = strchr(lbeg, '\n');
        llen = lend - lbeg;
    }

    if (lend == NULL) {
        // We can't complete the operation since the (last) header is
        // not terminated by two EOL characters.  Since it was already
        // processed (and modified with a null character), wipe it out
        // but leave a single "\n" in the buffer
        httpHeaders->clear();
        httpHeaders->push_back('\n');
        httpHeaders->push_back('\0');

        return size;
    }

    // Otherwise, we handled the header this round, so delete the vector
    consumed += (lend - str);
    httpHeaders->clear();
    headersSent = true;

    if (Malformed()
        || (((lbeg - &(*httpHeaders)[0]) != size) && (headersSent == false))) {
        Ns_Log(Debug, "PyWX: Malformed header: %s", lbeg);
        malformedHeader = true;
        return NS_ERROR;
    }

    Ns_Log(Debug, "PyWX: AOLbuffer consumed %d", consumed);

    return consumed;
}

//-----------------------------------------------------------------------------
//
//  int HandleHeader(void)
//
//  Handle the header line.  Keep this up-to-date with AOLserver.
//  Returns -1 on error, 0 for OK.
//
//-----------------------------------------------------------------------------
int
AOLbuffer::HandleHeader()
{
    char *field = &(*httpHeaders)[0];
    // Find the colon (and therefore the "field : value" string
    char *value = find(httpHeaders->begin(), httpHeaders->end(), ':');
    if (!value) {
        Ns_Log(Debug, "PyWX: Malformed header: %s", field);
        malformedHeader = true;
        return -1;
    }

    // Break the "field : value"
    *value = '\0';
    do {
        value++;
    } while (isspace(*value) && (value <= httpHeaders->end()));

    Ns_Log(Debug, "PyWX: Header is set; field=%s, value=%s", field, value);

    Ns_ConnSetHeaders(connection, field, value);
    if (!strcasecmp(field, "Location")) {
        Ns_Log(Debug, "PyWX: Aha -- redirected.");
        redirected = ((*value == '/') ? REDIRECT_INTERNAL : REDIRECT_EXTERNAL);
    }
    return NS_OK;
}

//-----------------------------------------------------------------------------
//
//  int sync()
//
//  Synchronize/flush the data in the buffer
//
//-----------------------------------------------------------------------------
int
AOLbuffer::sync()
{
    Ns_Log(Debug, "PyWX: AOLbuffer::sync");

    if (flushBuffer() == EOF) {
        return -1;
    }

    return 0;
}

//-----------------------------------------------------------------------------
//
//  int flushBuffer()
//
//  Flush the characters in the buffer to the output socket
//
//-----------------------------------------------------------------------------
int
AOLbuffer::flushBuffer()
{
    Ns_Log(Debug, "PyWX: AOLbuffer::flushBuffer");
    int num = pptr() - pbase();

    // Send output
    int rc = Ns_WriteConn(connection, writeBuffer, num);
    if (rc != 0) {
        Ns_Log(Debug, "PyWX: Error Condition: %d", rc);
        return EOF;
    }
    Ns_ConnFlushContent(connection);

    pbump(-num);			// Reset put pointer accordingly
    return num;
}

//-----------------------------------------------------------------------------
//
//  int overflow(int c)
//
//  Internal routine for writing the buffer contents when the buffer gets
//  filled up.
//
//-----------------------------------------------------------------------------
int
AOLbuffer::overflow(int c)
{
    Ns_Log(Debug, "PyWX: AOLbuffer::overflow");

    // If 'c' is not EOF, insert it into the buffer before dumping.
    if (c != EOF) {
        *pptr() = c;
        pbump(1);
    }

    // Flush the buffer
    if (flushBuffer() == NS_ERROR) {
        return NS_ERROR;
    }

    return c;
}

//-----------------------------------------------------------------------------
//
//  int underflow(void)
//
//  Internal routine for filling the AOLbuffer's internal read buffer
//  with data from the client.
//
//-----------------------------------------------------------------------------
int
AOLbuffer::underflow()
{
    Ns_Log(Debug, "PyWX: AOLbuffer::underflow --");
    if (ERROR()) {
        Ns_Log(Debug, "PyWX: AOLbuffer::underflow -- ERROR()");
        return NS_ERROR;
    }

    // Is the read position before "end-of-buffer"?
    if (gptr() < egptr()) {
        Ns_Log(Debug, "PyWX: AOLbuffer::underflow -- returning quickly");

        return *gptr();
    }

    // Process size of putback area:
    // - Use number of characters read
    // - But at most RB_OFFSET amount
    int numPutBack = gptr() - eback();
    if (numPutBack > RB_OFFSET) {
        numPutBack = RB_OFFSET;
    }

    // Copy up to RB_OFFSET characters previously read into
    // the putback buffer (area of first RB_OFFSET characters)
    memcpy(readBuffer + (RB_OFFSET - numPutBack), gptr() - numPutBack,
           numPutBack);

    // Now, read new data into the buffer
    readERROR = false;
    readEOF = false;

    int incoming = connection->contentLength;
    int amtRead = 0;
    if (incoming > 0) {
        int readAmt = ((incoming < (SMALLCHUNK - RB_OFFSET - 1)) ?
                       incoming : (SMALLCHUNK - RB_OFFSET - 1));
        amtRead = Ns_ConnRead(connection, readBuffer + RB_OFFSET, readAmt);
    }

    if (amtRead == NS_ERROR) {
        Ns_Log(Debug, "PyWX: AOLbuffer::underflow -- read error");

        // Set the Error Condition
        readERROR = true;
        return NS_ERROR;
    }

    if (amtRead == 0) {
        Ns_Log(Debug, "PyWX: AOLbuffer::underflow -- EOF");

        // Set the EOF Condition
        readEOF = true;
        return EOF;
    }

    // Reset buffer pointers
    setg(readBuffer + (RB_OFFSET - numPutBack), readBuffer + RB_OFFSET,
         readBuffer + RB_OFFSET + amtRead);

    Ns_Log(Debug, "PyWX: AOLbuffer::underflow -- normal return");

    return *gptr();
}

//-----------------------------------------------------------------------------
//
//  void Init(AOLbuffer* self, Ns_Conn* connection)
//
//  Python object initialization
//
//-----------------------------------------------------------------------------
void
AOLbuffer::Init()
{
    Ns_Log(Debug, "PyWX: AOLbuffer: Initializing values");
    headersSent = false;
    malformedHeader = false;
    httpHeaders = new Ns_Vector;
    redirected = 0;
    nph = false;
    setg(readBuffer + RB_OFFSET, readBuffer + RB_OFFSET,
         readBuffer + RB_OFFSET);
    setp(writeBuffer, writeBuffer + (TINYCHUNK - 1));

    readERROR = false;
    readEOF = false;
}

//-----------------------------------------------------------------------------
//
//  Destructors
//
//-----------------------------------------------------------------------------
AOLbuffer::~AOLbuffer()
{
    Ns_Log(Debug, "PyWX: AOLbuffer class destructor");
    sync();
    Ns_ConnFlushContent(connection);
    if (httpHeaders != NULL) {
        delete httpHeaders;
    }
}

/*
 * $Header: /cvsroot/PyWX/PyWX/AOLbuffer.cc,v 1.15 2000/08/26 01:24:57 mhagger Exp $
 */
