/* AOLbuffer source
 *
 * Copyright (C) 1999, 2000 Brent A. Fulgham <bfulgham@debian.org>
 * http://www.debian.org
 *
 * See also http://www.aolserver.com for details on the AOLserver
 * web server software
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 * 
 */

/*
 * The contents of this source file are based heavily on the following
 * sources:
 *
 * (*) Rob Mayoff's "tclpages" source code
 * (*) Lele Gaifax's "PyApache" source code
 * (*) Examples from the Python source distribution (especially the
 *     chapters on extended and embedding python)
 *
 * By the way, I tend to comment verbosely, so sorry if you are printing
 * this thing out.  But if I don't make notations here, I forget why I
 * did things and that just makes life harder in general.  Besides,
 * you're not paying for this so just enjoy. :-)
 * 
 */

#ifndef __AOLBUFFER__
#define __AOLBUFFER__

//#define Ns_Log(a,b) cerr << b << endl

/* 
 *==========================================================================
 *  This is a buffer class that allows us to glue AOLserver and Python
 *  together.  I can't take credit for the idea -- it's based on the
 *  ApacheBuff object from PyApache, though its implementation as a
 *  C++ class is my own.
 *
 *  Actually, I can't even claim all of the C++ as my own -- much of it
 *  is based on the work of the Python-C++ SIG, and specifically the
 *  CXX effort (and Gordon McMillan's SCXX classes).
 *==========================================================================
 */

#include <aolserver/ns.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <streambuf.h>
#include <vector>
#include <ctype.h>

const int REDIRECT_NONE = 0;
const int REDIRECT_INTERNAL = 1;
const int REDIRECT_EXTERNAL = 2;
const int SMALLCHUNK = 8192;
const int TINYCHUNK = 200;
const int HTTP_OK = 200;

// Forward declare AOLbuffer
class AOLbuffer;

//
//  Special Allocator (uses the Ns_Malloc methods to get its memory).
//
//  Implementation based on sample implementation from Nicolai Josuttis's
//  excellent "The C++ Standard Library".
//
template < class T > class SpecNsAlloc
{
public:
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    typedef T *pointer;
    typedef const T *const_pointer;
    typedef T & reference;
    typedef const T & const_reference;
    typedef T value_type;

    // Rebind allocator to type U
    template < class U > struct rebind
    {
        typedef SpecNsAlloc < U > other;
    };

    // Return address of values
    pointer address(reference value) const
    {
        return &value;
    }
    const_pointer address(const_reference value) const
    {
        return &value;
    }

    // These don't do anything
    SpecNsAlloc() throw() {}
    SpecNsAlloc(const SpecNsAlloc &) throw() {}
    template < class U > SpecNsAlloc(const SpecNsAlloc < U > &) throw() {}
    ~SpecNsAlloc() throw() {}

    // return maximum size of elements that can be allocated
    size_type max_size() const throw()
    {
        return LONG_MAX / sizeof(T);
    }

    // Allocate, but don't initialize "num" elements of type T
    pointer allocate(size_type num = SMALLCHUNK, const void *hint = 0) {
        return (pointer) (Ns_Malloc(num * sizeof(T)));
    }

    // Initialize elements of allocated storage 'p' with value 'value'
    void construct(pointer p, const T & value)
    {
        new((void *) p) T(value);
    }

    // Destroy elements of initialized storage 'p'
    void destroy(pointer p)
    {
        *p = NULL;
    }

    // Deallocate storage 'p' of deleted elements
    void deallocate(pointer p, size_type num)
    {
        Ns_Free((void *) p);
    }
};

// 
//  Return that all specializations of this allocator are interchangeable
//
template < class T1, class T2 >
bool operator == (const SpecNsAlloc < T1 > &,
                  const SpecNsAlloc < T2 > &) throw()
{
  return true;
}

template < class T1, class T2 >
bool operator != (const SpecNsAlloc < T1 > &,
                  const SpecNsAlloc < T2 > &) throw()
{
  return false;
}

//
//  Special vector definition -- uses the Ns_Alloc method to allocate
//  its memory
//
typedef vector < char, SpecNsAlloc < char > > Ns_Vector;

//========================================================================
//
//  The AOLbuffer class itself
//
//========================================================================
class AOLbuffer
    : public std::streambuf
{
protected:
    static const int RB_OFFSET = 20;
    Ns_Conn *connection;        	// The connection we are buffering
    char readBuffer[SMALLCHUNK];	// The buffered data from the client.
    char writeBuffer[TINYCHUNK];	// The buffered data from the server.

    // false until we parse and validate all the headers coming from
    // the script.  Then true:
    bool headersSent;

    bool malformedHeader;		// True for malformed header

    // Redirection status
    int redirected;

    // The header buffer.  If not empty, contains yet-to-be-parsed
    // headers:
    Ns_Vector *httpHeaders;

    bool nph;
    bool readEOF;
    bool readERROR;

    int underflow();
    int overflow(int c);

public:
    int flushBuffer();

    AOLbuffer(Ns_Conn * conn) : connection(conn)
    {
        Ns_Log(Debug, "PyWX: Constructing AOLbuffer");
    }
    ~AOLbuffer();

    void Init();
    int ParseHeaders(const char *str, int size);	// Parse HTTP Headers
    //void RedirectInternally(); // never defined
    int HandleHeader();
    int sync();

    // Inline methods
    bool IsNPH()
    { return nph; }

    void IsNPH(bool val)
    { nph = val; }

    bool EndOF()
    { return readEOF; }

    bool EMPTY()
    { return gptr() == egptr(); }

    bool ERROR()
    { return readERROR; }

    bool Malformed()
    { return malformedHeader; }

    bool HeaderSent()
    { return headersSent; }

    char *gReadData()
    { return gptr(); }

    char *gWriteData()
    { return pptr(); }

    // Return the position within the input stream
    int tell()
    { return gptr() - readBuffer; }
};

#endif // AOLBUFFER

/*
 * $Header: /cvsroot/PyWX/PyWX/AOLbuffer.hh,v 1.10 2000/08/26 01:24:57 mhagger Exp $
 */
