// For emacs==> -*- mode: C++; indent-tabs-mode: t; tab-width: 8; -*-

// $Id: ns_interps.cc,v 1.7 2000/08/26 01:24:57 mhagger Exp $

// Routines for maintaining a pool of Python interpreters: creating,
// preparing, and destroying them.

// See included file COPYRIGHT for copyright and license information.


#include "ns_interps.hh"


PyWX_InterpreterPool::PyWX_InterpreterPool(const char *pythonPath,
					   bool reuseInterpreters_)
    : interpStack(),
      interpMap(),
      pythonPagePath(0),
      reuseInterpreters(reuseInterpreters_),
      persistDict(PyDict_New())
{
    Ns_Log(Debug, "PyWX: Constructing PyWX_InterpreterPool.");
    if (pythonPath) {
	pythonPagePath = new char[strlen(pythonPath) + 1];
	assert(pythonPagePath);
	strcpy(pythonPagePath, pythonPath);
    }

    if (persistDict == NULL) {
	Ns_Log(Fatal, "PyWX: Can't create __persistdict__!");
	abort();
    }
}


PyWX_InterpreterPool::~PyWX_InterpreterPool()
{
    Ns_Log(Debug, "PyWX: Destroying PyWX_InterpreterPool.");

    delete [] pythonPagePath;
    Py_DECREF(persistDict);

    // Actually it would seem to make sense here to destroy any
    // interpreters that are in the pool.  However, they will be
    // destroyed anyway by the final Py_Finalize, so it is not
    // necessary.  Moreover, I haven't checked that this routine is
    // called only when it would be permissible to delete
    // interpreters.
}


int
PyWX_InterpreterPool::Create()
{
    Ns_Log(Debug, "PyWX: In PyWX_InterpreterPool.Create");

    PyThreadState *new_tstate = Py_NewInterpreter();
    if (! new_tstate) {
        Ns_Log(Error, "PyWX: Can't allocate thread.");
        return NS_ERROR;
    }

    // Add PageRoot path to the script's environment:
    if (pythonPagePath) {
        PyObject *path = PySys_GetObject("path"); // borrowed ref
        PyList_Insert(path, 0, TempPyString(pythonPagePath));
    }

    // Fool the interpreter into thinking it is executing these
    // scripts as a "main".  This is based on the PyApache code.
    PyObject *mainmodule = PyImport_AddModule("__main__"); // borrowed ref
    if (! mainmodule) {
        Ns_Log(Fatal, "PyWX: Cannot add __main__ module");
        Discard(); // error exit
        return NS_ERROR;
    }

    // Provide __persistdict__:
    if (SetupPersist() != NS_OK) {
	Discard();
	return NS_ERROR;
    }

    return NS_OK;
}


int
PyWX_InterpreterPool::Acquire()
{
    Ns_Log(Debug, "PyWX: In PyWX_InterpreterPool.Acquire");

    if (interpStack.empty()) {
        // no interpreters on stack; create a new one:
        Ns_Log(Debug, "PyWX: No free interpreters - creating new one.");
        return Create();
    }
    else {
        // pull one off the stack
        Ns_Log(Debug, "PyWX: Popping interpreter off stack");
        PyThreadState *new_tstate = interpStack.top();
        interpStack.pop();
        PyThreadState_Swap(new_tstate);

        // __main__ is added when an interpreter is recycled, so we
        // don't need to add it again.

	// Provide __persistdict__:
	if (SetupPersist() != NS_OK) {
	    Discard();
	    return NS_ERROR;
	}

	return NS_OK;
    }
}


void
PyWX_InterpreterPool::Recycle()
{
    if (! reuseInterpreters) {
        Discard();
    }
    else {
        // clear the interpreter, prepare it for reuse, and store
        // it on interpStack.

        PyThreadState *tstate = PyThreadState_Get();
        PyThreadState_Clear(tstate);   // clear stuff from thread state

        // Also, clean up the __main__ dictionary, adding only __builtins__
        PyObject *mainmodule = PyImport_AddModule("__main__");
        if (mainmodule == NULL) {
            Py_FatalError("can't create __main__ module");
        }

        PyObject *maindict = PyModule_GetDict(mainmodule);
        PyDict_Clear(maindict);

        PyObject *bimod = PyImport_ImportModule("__builtin__");
        if (
                bimod == NULL ||
                PyDict_SetItemString(maindict, "__builtins__", bimod) != 0
                ) {
            Py_FatalError("can't add __builtins__ to __main__");
            //Py_DECREF(bimod); // @@MRH Line never reached
        }

        PyThreadState_Swap(NULL); // swap out
        interpStack.push(tstate); // push onto stack
    }
}


void
PyWX_InterpreterPool::Discard()
{
    Py_EndInterpreter(PyThreadState_Get());
}


void
PyWX_InterpreterPool::Reserve(Ns_Conn *conn)
{
    // Store the interpreter in the map keyed by connection id:
    interpMap[conn] = PyThreadState_Swap(NULL);
}


bool
PyWX_InterpreterPool::Retrieve(Ns_Conn *conn)
{
    // Look for an existing interpreter associated with this connection:
    map < Ns_Conn *, PyThreadState * >::iterator pyeval =
        interpMap.find(conn);
    if (pyeval == interpMap.end()) {
        // No existing interpreter for this connection.
        Ns_Log(Debug, "PyWX: Didn't find interpreter in pyinterp map");
        return false;
    }
    else {
        PyThreadState *new_tstate = pyeval->second;
	Ns_Log(Debug, "PyWX: Using existing pyinterp %x.", new_tstate);
        // Erase the interpreter from the map:
        interpMap.erase(pyeval);
	// Swap it in:
	PyThreadState *old_tstate = PyThreadState_Swap(new_tstate);
        if (old_tstate != NULL) {
            Ns_Log(Fatal, "PyWX: Old thread state was not empty!");
            abort();
        }
	return true;
    }
}


void
PyWX_InterpreterPool::Cleanup(Ns_Conn *conn)
{
    map < Ns_Conn *, PyThreadState * >::iterator pyeval =
	interpMap.find(conn);

    // is there a valid interp?
    if (pyeval != interpMap.end()) {
	Ns_Log(Debug, "PyWX: Cleaning up. --");
	PyThreadState *tstate = pyeval->second;
	interpMap.erase(pyeval); // Zap the interpreter from the map
	PyThreadState *oldstate = PyThreadState_Swap(tstate); // swap it in
	assert(oldstate == NULL);
	Recycle(); // Leaves thread state in NULL.
	Ns_Log(Debug, "PyWX: -- Cleaned up.");
    }
}


// Add __persistdict__ to the __builtins__ module.
int PyWX_InterpreterPool::SetupPersist()
{
    PyObject *builtins = PyEval_GetBuiltins(); // borrowed ref

    if (PyDict_SetItemString(builtins, "__persistdict__", persistDict)) {
	Ns_Log(Error, "PyWX: Cannot add __persistdict__ to __builtins__.");
	return NS_ERROR;
    }

    return NS_OK;
}


