// For emacs==> -*- mode: C++; indent-tabs-mode: t; tab-width: 8; -*-

// $Id: ns_interps.hh,v 1.6.4.1 2000/08/28 16:34:55 mhagger Exp $

// See included file COPYRIGHT for copyright and license information.


#ifndef __NS_INTERPS__
#define __NS_INTERPS__

#include <map>
#include <stack>

#include "ns_pywx.hh"


// Maintain a pool of the active python interpreters, and create and
// destroy interpreters as needed.  Every interpreter that is not in
// use is held here.  Possible states of an interpreter:
//
// 1. In use by an active thread.  Such interpreters are not reflected
//    in this data structure.
//
// 2. No longer in use, but able to be reused by a new connection (or
//    connectionless).  Such interpreters are held in interpStack.
//
// 3. Not currently in use, but may possibly need to be re-used by the
//    same connection.  Such interpreters are held in interpMap, which
//    is a map indexed by (Ns_Conn *).  This doesn't seem like a good
//    key, but in fact when a connection is destroyed the Cleanup()
//    method should be called (deleting the interpreter from the map)
//    before the Ns_Conn has a chance to be reused.
//
// Note that these methods should only be called while the Python
// global interpreter lock is held.  This is not only for Python's
// benefit, but also because this class doesn't do any other
// serialization internally.
//
// There is a potential problem here: if code calls Create() then
// Recycle() repeatedly, an unlimited number of interpreters could be
// put on the stack but never freed.  Therefore, if you need to use
// Create(), you should probably pair it with Discard().  ### We
// should probably implement an upper limit to the stack size and
// Discard interpreters after that.

class PyWX_InterpreterPool
{
public:
    PyWX_InterpreterPool(const char *pythonPath,
			 bool reuseInterpreters_);
    ~PyWX_InterpreterPool();

    // Create a brand-new interpreter and swap it in.
    int Create();

    // Acquire an unused interpreter from the pool if available;
    // otherwise create a new one.  In either case, swap it in.
    int Acquire();

    // Swap out the interpreter and clear it, and store it in the
    // interpreter stack so that it can be recycled:
    void Recycle();

    // Swap out the interpreter and destroy it (prevent reuse):
    void Discard();

    // Swap out the interpreter but maintain it in the interpreter map
    // in case it is needed again by the specified connection:
    void Reserve(Ns_Conn *conn);

    // If an interpreter was already in use for the specified
    // connection, fetch it, swap it in, and return true.  Otherwise
    // return false.
    bool Retrieve(Ns_Conn *conn);

    // If there is an interpreter Reserved by the specified
    // connection, clean it out and recycle it.
    void Cleanup(Ns_Conn *conn);

private:
    int SetupPersist();

private:
    stack<PyThreadState *> interpStack;
    map<Ns_Conn *, PyThreadState *> interpMap;

    char *pythonPagePath; // path to add to sys.path in new interpreters
    bool reuseInterpreters;

    // A server-wide, persistent dictionary that all scripts will have
    // access to.  Scripts will receive it as __persistdict__ in their
    // global namespace.
    PyObject *persistDict;
};


#endif __NS_INTERPS__

