// For emacs==> -*- mode: C++; indent-tabs-mode: t; tab-width: 8; -*-

// $Id: ns_pywx.cc,v 1.66.4.1 2000/08/29 14:18:47 mhagger Exp $

// Run a Python page (possibly from a file cache).


/* PyWX source
 *
 * Copyright (C) 1999 Brent A. Fulgham <bfulgham@debian.org>
 * http://www.debian.org
 *
 * See also http://www.aolserver.com for details on the AOLserver
 * web server software
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 * 
 */

/*
 * The contents of this source file are based heavily on the following
 * sources:
 *
 * (*) Rob Mayoff's "tclpages" source code
 * (*) Lele Gaifax's "PyApache" source code
 * (*) Examples from the Python source distribution (especially the
 *     chapters on extended and embedding python)
 *
 * However, the code itself has since been extensively rewritten many
 * times.
 *
 * This source file is just the first step in Python world domination.
 * If this unit is successful, it will mean the begining of my secret
 * plot to replace the AOLserver Tcl base with Python.  Mwahahahaha.
 *
 */

// Don't callback functions technically need C linkage?

// Python header files
#include <sys/stat.h>

#include "ns_pywx.hh"
#include "PyWX_buffer.hh"

#include "ns_interps.hh"

int Ns_ModuleVersion = 1;
const char *revision = "0.6"; // also set in ns_setup.py


// Global variables used within this file:

// A pool of recyclable interpreters:
static PyWX_InterpreterPool *interpPool;

static PyThreadState *parentInterpreter;

// Should python's stdio be attached to a PyWX_buffer when there is a
// connection?
static bool allowAttachStdio;

// Should the interpreter's environment variables and sys.argv be set
// up like those of a CGI script?  ### Currently, setting this to true
// causes a memory leak.
static bool setupCGIEnv;


#define THREAD_DEBUG 0
#if THREAD_DEBUG

// Install debugging versions of lock routines.

#define _PyEval_ReleaseLock() \
do { \
    PyEval_ReleaseLock(); \
    Ns_Log(Debug, "PyWXb: release lock at %s:%d", __FILE__, __LINE__); \
} while (0)


#define _PyEval_AcquireLock() \
do { \
    Ns_Log(Debug, "PyWXb: trying to acquire lock at %s:%d", \
           __FILE__, __LINE__); \
    PyEval_AcquireLock(); \
    Ns_Log(Debug, "PyWXb: acquire lock at %s:%d", __FILE__, __LINE__); \
} while (0)


#else /* THREAD_DEBUG */

#define _PyEval_AcquireLock() PyEval_AcquireLock()
#define _PyEval_ReleaseLock() PyEval_ReleaseLock()

#endif /* THREAD_DEBUG */


// This is the Python script used to run a particular page from PageGet.
// Ideally, we want to cache this.  For now, be simple!
static char runPythonScriptText[] = "import pywx; pywx.fulfill_request()";


// Setup the Python Interpreter environment to be like that of a CGI
// script by calling ns_setup.setup_cgi_env(filename).
static int
SetupPythonCGIEnv(char *filename)
{
    TempPyObject setup_module = PyImport_ImportModule("ns_setup"); // new ref
    if (! setup_module) {
	Ns_Log(Error, "PyWX: Could not import ns_setup!");
	PyErr_Print();
	return NS_ERROR;
    }

    // can't fail:
    PyObject *setup_dict = PyModule_GetDict(setup_module); // borrowed ref
    PyObject *setup_fn =
	PyDict_GetItemString(setup_dict, "setup_cgi_env"); // borrowed ref
    if (! setup_fn) {
	Ns_Log(Error, "PyWX: Could not find ns_setup.setup_cgi_env!");
	PyErr_Print();
	return NS_ERROR;
    }

    TempPyObject args = Py_BuildValue("(s)", filename); // new reference
    if (! args) {
	Ns_Log(Error, "PyWX: Couldn't create setup_cgi_env argument list!");
	PyErr_Print();
	return NS_ERROR;
    }

    Ns_Log(Debug, "PyWX: Calling ns_setup.setup_cgi_env()");
    Py_INCREF(setup_fn); // prevent premature destruction
    // Reference-count neutral WRT args; returns new ref:
    TempPyObject result = PyObject_CallObject(setup_fn, args);
    Py_DECREF(setup_fn);

    if (result == NULL) {
	Ns_Log(Error, "PyWX: ns_setup.setup_cgi_env() failed!");
	PyErr_Print();
	return NS_ERROR;
    }
    return NS_OK;
}


static void
SetupCurrPythonIOBuffer(Ns_Conn *conn, int isNPH)
{
    // Build an AOLserver/Python bridge buffer and bind stdin/stdout to it.
    PyWX_buffer *buff =
	(PyWX_buffer *)PyObject_NEW(PyWX_buffer, &PyWXbuffType);
    buff->Init(buff, conn);	// Initialize values

    if (isNPH) {
	buff->IsNPH(true);
    }

    PySys_SetObject("stdin", (PyObject *)buff);
    PySys_SetObject("stdout", (PyObject *)buff);

    // Decrement reference -- it's now "owned" by the interpreter and
    // will be destroyed when the interpreter is done.
    Py_DECREF((PyObject *) buff);
}


// PyWX_Cleanup: registered to run at the end of each connection.
// Cleans up Python interpreters.
static void
PyWX_Cleanup(void *ignored, Ns_Conn *conn)
{
    // Since PyWX_Cleanup is run at the end of the connection, this
    // function can serve to wipe the namespace, etc., allowing ADP
    // and Tcl pages to do multiple calls to evaluate Python
    // statements, which will share the same interpreter state.

    _PyEval_AcquireLock();
    interpPool->Cleanup(conn);
    _PyEval_ReleaseLock();
}


// Call the 'flush' method on the 'stdout' object.  Should only be called
// if there's a current connection, obviously.
static void PyWX_FlushStdout()
{
    PyObject *out = PySys_GetObject("stdout"); // borrowed ref
    if (! out) {
	Ns_Log(Error, "PyWX: Can't get sys.stdout.");
	PyErr_Print();
	return;
    }
    Py_INCREF(out); // Prevent premature death
    PyObject_CallMethod(out, "flush", NULL);
    Py_DECREF(out);
}


// flags for PyWX_PyExecute:
// The NO_CONN flag is gone because that can be determined from conn==NULL.
#define EVAL               0x01	      // use 'Py_eval_input' == 'eval'
#define NO_ATTACH_STDIO    0x03       // do not attach stdio to AOLServer

// PyWX_PyExecute: The workhorse function and main python entry point.
//
// Arguments:
//	sourceName : what to put in Python error messages
//	pythonText : python code to execute
//	conn       : (Ns_Conn *) for the connection; also used as a unique ID
//		     used to look up things on a per-connection basis.
//	setupFlags : how to set up the environment.
//	interp     : Tcl interpreter to which error text or results text
//                   are returned, or NULL if results should be discarded.
//
// Return value: false if fail, true if succeed.
static bool PyWX_PyExecute(char *sourceName, char pythonText[],
			   Ns_Conn *conn, int setupFlags,
			   Tcl_Interp *interp)
{
    // initialize various settings based on flags:

    bool attachStdio = (conn != NULL
			&& allowAttachStdio
			&& ! (setupFlags & NO_ATTACH_STDIO));

    int startType = (setupFlags & EVAL) ? Py_eval_input : Py_file_input;

    // Try to use previous interpreter, to preserve state across
    // multiple 'ns_python' statements:
    if (conn && interpPool->Retrieve(conn)) {
	// just reuse the old interpreter--done!
	Ns_Log(Debug, "PyWX: Reusing existing interpreter.");
    }
    else {
	// We need to create a new python interpreter for this request.
	Ns_Log(Debug, "PyWX: No existing pyinterp for this connection");

	if (interpPool->Acquire() == NS_ERROR) {
	    Ns_Log(Error, "PyWX: Couldn't acquire Python interpreter.");
	    return false;
	}

	if (conn && setupCGIEnv) {
	    // Set up interpreter:
	    if (SetupPythonCGIEnv(sourceName) != NS_OK) {
		Ns_Log(Error, "PyWX: Can't setup interpreter");
		interpPool->Discard();	// error exit
		return false;
	    }
	}

	// Attach IO <--> AOLserver
	if (attachStdio) {
	    SetupCurrPythonIOBuffer(conn, 0);
	}
    }

    Ns_Log(Debug, "PyWX: args: \"\"\"%s\"\"\", %s, %d",
	   pythonText, sourceName, setupFlags);

    PyCodeObject *code =
	(PyCodeObject *) Py_CompileString(pythonText, sourceName, startType);

    if (! code) {
	PyObject *error = PyErr_Occurred();
	const char *msg = NULL;
	if (error != NULL) {
	    if (!PyErr_GivenExceptionMatches(error, PyExc_SystemExit)) {
		Ns_Log(Error,
		       "PyWX: Compilation of script raised an unhandled "
		       "exception (PyWX_PyExecute)");
		PyErr_Print();
		msg = "ns_python eval: couldn't compile";
	    }
	    else {
		msg = "ns_python eval: unknown error in compilation";
	    }
	}
	else {
	    msg = "ns_python eval: cannot compile statement";
	}
	if (interp) {
	    Tcl_AppendResult(interp, msg, NULL);
	}
	// Regardless of what the error was, we want to abort
	interpPool->Discard();	// error exit
	return false;
    }

    // OK.  Now we've got the code; let's run it.

    // dictionary of __main__ (used as global dict):
    PyObject *mainmodule = PyImport_AddModule("__main__"); // borrowed ref
    if (! mainmodule) {
	// This shouldn't happen, since it was added successfully in
	// interpPool->Acquire().
	Ns_Log(Error, "PyWX: Fatal error.  Cannot add __main__ module");
	interpPool->Discard();	// error exit
	return false;
    }

    PyObject *maindict = PyModule_GetDict(mainmodule); // borrowed ref
    TempPyObject result = PyEval_EvalCode(code, maindict, maindict);

    Py_DECREF(code);		// Code is no longer of use.

    // Do result/error handling.
    bool retval = true;
    if (result.get_ptr() == NULL) {
	// failure
	retval = false;
	PyObject *error = PyErr_Occurred();
	// should we print something out?
	if (error) {
	    if (!PyErr_GivenExceptionMatches(error, PyExc_SystemExit)) {
		Ns_Log(Error, "PyWX: Script raised an unhandled exception.");
		PyErr_Print();
		if (interp) {
		    Tcl_AppendResult(interp,
				     "PyExecute: runtime failure", NULL);
		}
	    }
	}
	else {
	    // error == NULL
	    if (interp) {
		Tcl_AppendResult(interp,
				 "PyWX_PyExecute: unknown runtime failure",
				 NULL);
	    }
	}

	// Even if it's an error, ReleaseInterpreter will be called
	// within PyWX_Cleanup.
    }
    else {
	// If we're in 'EVAL' mode and 'result' is not None, convert
	// it to a string and return it to Tcl:
	if (interp
	    && (setupFlags & EVAL)
	    && result.get_ptr() != Py_None) {
	    TempPyObject resultString = PyObject_Str(result.get_ptr());
	    if (resultString.get_ptr() != NULL
		&& PyString_Check(resultString.get_ptr())) {
		Tcl_AppendResult(interp,
				 PyString_AS_STRING(resultString.get_ptr()),
				 NULL);
	    }
	}
    }

    // Flush any output.
    if (attachStdio) {
	PyWX_FlushStdout();
    }

    if (conn) {
	// keep interpreter, associated with connection, in case
	// needed again:
	interpPool->Reserve(conn);
    }
    else {
	// recycle interpreter
	interpPool->Recycle();
    }
    return retval;
}


// Run a Python file.  Return NS_OK for success or NS_ERROR for failure.
static int
PyWX_PageGet(void *ignored, Ns_Conn *conn)
{
    // There shouldn't be any connectionless requests here, right?
    assert(conn != 0);

    Ns_Log(Debug, "PyWX_PageGet: url %s", conn->request->url);

    Ns_DString filename;
    Ns_DStringInit(&filename);
    if (Ns_UrlToFile(&filename, nsServer, conn->request->url) != NS_OK) {
	Ns_DStringAppend(&filename, "<no filename available>");
    }

    _PyEval_AcquireLock();
    bool retval = PyWX_PyExecute(Ns_DStringValue(&filename),
				  runPythonScriptText, conn, 0, NULL);
    _PyEval_ReleaseLock();

    Ns_DStringFree(&filename);

    // Now, close connection in case the process lingers...  This will
    // automatically flush the buffer.
    Ns_ConnClose(conn);

    return retval ? NS_OK : NS_ERROR;
}


// Register some handlers for various types of HTTP methods.
static void
RegisterMap(char *hServer, char *map)
{
    Ns_Log(Notice, "PyWX: registering map \"%s\"", map);
    Ns_RegisterRequest(hServer, "GET", map, PyWX_PageGet, NULL, NULL, 0);
    Ns_RegisterRequest(hServer, "HEAD", map, PyWX_PageGet, NULL, NULL, 0);
    Ns_RegisterRequest(hServer, "POST", map, PyWX_PageGet, NULL, NULL, 0);
}


// Associate the module with file extensions from the `map' directive
// in the module's configuration section.
static void
RegisterMaps(char *hServer, char *moduleConfigPath)
{
    Ns_Set *set = Ns_ConfigGetSection(moduleConfigPath);
    if (set != NULL) {
	int setSize = Ns_SetSize(set);
	for (int i = 0; i < setSize; i++) {
	    if (STRIEQ(Ns_SetKey(set, i), "map")) {
		RegisterMap(hServer, Ns_SetValue(set, i));
	    }
	}
    }
}


// Various destruction/cleanup for when PyWX is unloaded from the server.
static void
PyWX_TerminateModule(void *context)
{
    Ns_Log(Notice, "PyWX: Terminating.");

    _PyEval_AcquireLock();
    delete interpPool;
    interpPool = 0;
    PyThreadState_Swap(parentInterpreter);
    _PyEval_ReleaseLock();

    Py_Finalize();
}


// Called from Tcl as 'ns_python eval'.  Returns TCL_ERROR + error
// string for wrong # args, no conn; returns TCL_ERROR on anything it
// doesn't understand.
static int
Ns_PythonEvalCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
    int flags = 0;

    // Check # of args:
    if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # of args: should be \"",
			 argv[0], " eval|exec \"pythonText\" ", NULL);
	return TCL_ERROR;
    }

    /* Set up the interpreter environment flags: eval or exec? */

    if (!strcmp(argv[1], "eval")) {
	flags = flags | EVAL;
    }
    else if (!strcmp(argv[1], "exec")) {
	// default flags
    }
    else {
	Tcl_AppendResult(interp, "unavailable argument \"",
			 argv[1], " should be 'eval' or 'exec'", NULL);
	return TCL_ERROR;
    }

    /* Pass through the text to execute */
    char *pythonText = argv[2];

    /* Get the script name */

    // CTB note: 'info script' can't cause a python call or we'll have
    // to release lock here.  "info script" returns the name of the
    // innermost Tcl script being executed, or "" otherwise.
    char scriptName[500] = "??";	// @CTB -- don't statically allocate!
    if (Tcl_Eval(interp, "info script") == TCL_OK) {
	// It turns out not to be easy to get the line number from within
	// Tcl...  -- CTB
	// @MRH Possible buffer overflow here:
	sprintf(scriptName, "Tcl script %s (in 'ns_python %s')",
		argv[1], interp->result);
    }

    /* Run the Python code. */
    _PyEval_AcquireLock();
    bool retval = PyWX_PyExecute(scriptName, pythonText, Ns_GetConn(),
				 flags, interp);
    _PyEval_ReleaseLock();

    return retval ? TCL_OK : TCL_ERROR;
}


// Register Python-specific commands with Tcl interpreter.  Returns
// TCL_OK.
static int
RegisterTclPythonFunctions(Tcl_Interp *interp, void *dummy)
{
    Tcl_CreateCommand(interp, "ns_python", Ns_PythonEvalCmd, NULL, NULL);
    return TCL_OK;
}


// Setup the Python Interpreter's Environment that will be effective
// globally.  Whatever is set up here will be copied to other new
// interpreters.
// 
// - Add PageRoot to sys.path.
// - Import ns_setup.py.
// - Run ns_setup.setup_global_interpreter().
static int
SetupGlobalPythonEnv(const char *pythonPagePath)
{
    // BAF -- Don't acquire global interp lock here.  This routine is
    // called by the main module entry which already holds the lock at
    // this point.

    PyObject *path = PySys_GetObject("path"); // borrowed ref
    if (! path) {
	Ns_Log(Error, "PyWX: Cannot get sys.path");
	return NS_ERROR;
    }
    if (pythonPagePath) {
	Ns_Log(Debug, "PyWX: Adding %s to sys.path", pythonPagePath);
	PyList_Insert(path, 0, TempPyString(pythonPagePath));
    }
    else {
	Ns_Log(Debug,
	       "PyWX: PageRoot isn't set; I hope we can find ns_setup.py...");
    }
    Ns_Log(Debug, "PyWX: importing ns_setup");
    TempPyObject setup_module = PyImport_ImportModule("ns_setup"); // new ref
    if (! setup_module) {
	Ns_Log(Error, "PyWX: Could not import ns_setup!");
	PyErr_Print();
	return NS_ERROR;
    }

    PyObject *setup_dict = PyModule_GetDict(setup_module); // borrowed ref

    // borrowed:
    PyObject *setup_fn =
	PyDict_GetItemString(setup_dict, "setup_global_interpreter");

    if (! setup_fn) {
	Ns_Log(Error, "PyWX: Couldn't find ns_setup.setup_global_interpreter");
	PyErr_Print();
	return NS_ERROR;
    }

    Ns_Log(Debug, "PyWX: Calling ns_setup.setup_global_interpreter");
    Py_INCREF(setup_fn); // prevent premature destruction
    TempPyObject result = PyObject_CallObject(setup_fn, NULL); // new ref
    Py_DECREF(setup_fn);

    if (result) {
	Ns_Log(Debug,
	       "PyWX: ns_setup.setup_global_interpreter "
	       "returned successfully.");
    }
    else {
	Ns_Log(Error, "PyWX: ns_setup.setup_global_interpreter() failed!");
	PyErr_Print();
	return NS_ERROR;
    }

    return NS_OK;
}


static bool GetBoolOption(char *configpath, char *name, bool default_value)
{
    int value;
    int rc = Ns_ConfigGetBool(configpath, name, &value);

    if (rc == NS_FALSE) {
	Ns_Log(Notice, "PyWX: %s not set; defaults to %s.",
	       name, default_value ? "on" : "off");
	return default_value;
    }	    
    else if (value) {
	Ns_Log(Notice, "PyWX: %s set to on.", name);
	return true;
    }
    else {
	Ns_Log(Notice, "PyWX: %s set to off.", name);
	return false;
    }
}


// Initialization function called at server start up.  Returns NS_OK
// or NS_ERROR.  May create file cache.
int
Ns_ModuleInit(char *hServer, char *hModule)
{
    // Note:  Don't call Py_SetProgramName() because we want the
    // default search path to be to the regular Python libraries
    // and modules.  We can append others to it later

    Py_Initialize();
    Ns_Log(Debug, "PyWX: Initializing threads");
    PyEval_InitThreads();

    char *serverConfigPath = Ns_ConfigGetPath(hServer, NULL, NULL);
    char *moduleConfigPath = Ns_ConfigGetPath(hServer, hModule, NULL);

    RegisterMaps(hServer, moduleConfigPath);
    Ns_RegisterShutdown(PyWX_TerminateModule, NULL);

    // Register Tcl commands at interp start:
    Ns_Log(Notice, "PyWX: Registering PyWX Python interface functions");

    Ns_TclRegisterAtCreate(RegisterTclPythonFunctions, NULL);

    // Register a "cleanup" function to be called at the end of
    // each page:
    Ns_Log(Notice, "PyWX: Registering PyWX_Cleanup function");
    Ns_RegisterCleanup(PyWX_Cleanup, NULL);

    // set reuseInterpreters from the configuration value:
    bool reuseInterpreters = GetBoolOption(
	    moduleConfigPath, "ReuseInterpreters", true);

    // set allowAttachStdio from the configuration value:
    allowAttachStdio = GetBoolOption(
	    moduleConfigPath, "AttachStdio", true);

    // set setupCGIEnv from the configuration value:
    setupCGIEnv = GetBoolOption(moduleConfigPath, "SetupCGIEnv", true);

    // Read the PyWX PageRoot parameter from the configuration file:
    char *pythonPagePath =
	Ns_ConfigGetValue("ns/environment/python", "PageRoot");
    if (pythonPagePath != NULL && pythonPagePath[0] == '\0') {
	pythonPagePath = NULL;
    }

    // initialize the interpreter pool:
    interpPool = new PyWX_InterpreterPool(pythonPagePath, reuseInterpreters);

    // Set up the python environment.
    if (SetupGlobalPythonEnv(pythonPagePath) == NS_ERROR) {
	return NS_ERROR;
    }

    if (allowAttachStdio) {
	initPyWXbuffer();
	// This should always be true; I don't know why it used to be
	// set explicitly:
	assert(PyWXbuffType.ob_type == &PyType_Type);
    }

    parentInterpreter = PyThreadState_Swap(NULL);

    _PyEval_ReleaseLock();

    return NS_OK;
}


